#!/usr/bin/python

from ConfigParser import ConfigParser
import datetime
import errno
import getopt
import json
import logging
import math
import os
import pwd
import sqlite3
import sys
from threading import Thread
from threading import current_thread
import time

data_container = {}
container_json = {}
user_map = {}
site_map = {}

# set config initialization
config = ConfigParser()

# parse existing file
config.read('/usr/local/safesquid/api/config.ini')

#date
this_year = time.strftime("%Y")
this_month = time.strftime("%m")
this_day = time.strftime("%d")
this_hrs = time.strftime("%H")

# read values from a section
log_path = config.get('log_info', 'log_path')
log_file = config.get('log_info', 'log_file')
db_resource = config.get('database', 'resource_db')
report_src = config.get('report_path', 'report_dir')
user = config.get('permission', 'user')
group = config.get('permission', 'group')
from_date = ''
to_date = ''
report_path = ''
report_type = ''

# logs to be written to file
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# create a file handler
log_file_path = log_path + log_file
handler = logging.FileHandler(log_file_path)
handler.setLevel(logging.INFO)

# create a logging format
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

# add the handlers to the logger
logger.addHandler(handler)

def date_arguments():    
    global from_date, to_date, report_path, report_type
    date_gen = ''
    date_split = []
    def usage():
        print 'Usage : ./query_builder.py                                                -- Generate for last one hour'
        print '        ./query_builder.py -s <start date> -e <end date> -t <day|hrs|mon> -- generate between time range'        
    
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hs:e:t:d:")
    except getopt.GetoptError as err:        
        print str(err)
        usage()
        sys.exit(2)
        
    if (len(opts) == 0):        
        lastHourDateTime = datetime.datetime.now() - datetime.timedelta(hours=1)        
        this_hour = time.strftime("%d.%m.%Y %H:00:00") 
        this_lasthour = lastHourDateTime.strftime("%d.%m.%Y %H:00:00")
        last_hrs = lastHourDateTime.strftime("%H")
        
        from_date = str(epoch_time(this_lasthour))
        to_date = str(epoch_time(this_hour))        
        report_path = report_src + '/' + this_year + '/' + this_month + '/' + this_day + '/' + last_hrs
        
    else:
        logger.info('total arg passed: %s', len(opts))
        for opt, arg in opts:
            if opt == '-h':
                usage()
                sys.exit()
            elif opt in ("-s"):
                from_date = str(epoch_time(arg))
            elif opt in ("-e"):
                to_date = str(epoch_time(arg)) 
            elif opt in ("-t"):
                report_type = arg
            elif opt in ("-d"):
                date_gen = arg
        
        report_path = report_src + '/' + this_year + '/' + this_month + '/' + this_day + '/' + this_hrs  
        
        if report_type:
            that_day = time.strftime(date_gen + " 00:00:00") 
            that_day_end = time.strftime(date_gen + " 23:59:59")
            
            from_date = str(epoch_time(that_day))
            to_date = str(epoch_time(that_day_end)) 
            
            date_split = date_gen.split('.')
            
            report_path = report_src + '/' + date_split[2] + '/' + date_split[1] + '/' + date_split[0]  
        
    logger.info('Generating report for [%s:%s] in %s', from_date, to_date, report_path)
    
#epoch time conversion
def epoch_time(date_time=None):
    if date_time is None:
        date_time = time.strftime("%d.%m.%Y %H:%M:%S")

    pattern = '%d.%m.%Y %H:%M:%S'
    epoch = int(time.mktime(time.strptime(date_time, pattern)))
    return epoch
        
#write to files
def write_file(file_name, data):    
    with open(file_name, 'wb') as x_file:
        x_file.write(data)

#query handler
class GetQueryThread(Thread):
    def __init__(self, url):
        self.url = url[0]                
        self.tag = url[1]
        super(GetQueryThread, self).__init__()

    def run(self):
        start_s = time.time()
        thread_id = current_thread().name
        conn = sqlite3.connect(db_resource)
        
        try:
            conn.row_factory = sqlite3.Row
            curs = conn.cursor()
            logger.info('[%s] Quering for [%s] : %s', thread_id, self.tag, self.url)
            curs.execute(self.url)
            recs = curs.fetchall()
        except sqlite3.Error as er:
            logger.error('[%s] error: %s', thread_id, er.message)

        conn.close()

        rows = [dict(rec) for rec in recs]

        logger.info('[%s] Complted query in : %s', thread_id, (time.time()-start_s))        
        
        #write_file(report_path + "/" + self.tag, json.dumps(rows))
        data_container[self.tag] = rows

def check_directory(path):
    if not os.path.exists(path):
        try:
            uid, gid = pwd.getpwnam(user).pw_uid, pwd.getpwnam(group).pw_uid
            logger.info('creating directory %s with [%s:%s]', path, user, group)
            os.makedirs(path)
            os.chown(path, uid, gid)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                logger.error('creating directory %s failed', path)
                raise

def convert_size(size_bytes):
    if (size_bytes == 0):
        return '0B'
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return '%s %s' % (s, size_name[i])

def getSize(filename):
    st = os.stat(filename)
    return convert_size(st.st_size)
    
def find_key(input_dict, value):    
    return next((k for k, v in input_dict.items() if v == value), None)

def get_responses():    
    global_1 = "select count(DISTINCT L) as bu,count(D) as rq,sum(G) as tb from ( select L,D,G from main order by D ) where D between " + from_date + " and " + to_date + " ;"
    global_2 = "select count(DISTINCT N) as bls from (select N,L,D from main where F=403 ) where D between " + from_date + " and " + to_date + " ;"
    global_3 = "select count(DISTINCT N) as brs from (select N,L,D from main where F NOT IN(503,500,407,403) ) where D between " + from_date + " and " + to_date + " ;"
    userlist = "select DISTINCT users_master.A as users from (select L,D from main where D between " + from_date + " and " + to_date + "),users_master where L=users_master.B order by users;"
    sitelist = "select url_master.A as urls from (select N,D from main where D between " + from_date + " and " + to_date + "),url_master where N=url_master.B group by N order by urls;"
    user_details_1 = "select users_master.A as users, rq, bt,ul,dw, bw from (select L as user, count(D) as rq, sum(E) as bt, sum(H) as ul, sum(I) as dw, sum(G) as bw from (select D,L,G,E,H,I from main order by D) where D between " + from_date + " and " + to_date + " group by L order by bw DESC),users_master where user=users_master.B;"
    user_details_2 = "select users_master.A as users, dn from(select count(DISTINCT N) as dn, L from main where F=403 and D between " + from_date + " and " + to_date + " group by L),users_master where L=users_master.B;"
    user_details_3 = "select users_master.A as users, br from(select count(DISTINCT N) as br,L from main where F NOT IN (503,500,407,403) and D between " + from_date + " and " + to_date + " group by L),users_master where L=users_master.B;"
    user_info_1 = "select interval, users_master.A as users, rq, bw from (select strftime('%d',datetime(D, 'unixepoch', 'localtime')) as interval, L as user, count(L) as rq,sum(G) as bw from (select D,L,G from main order by D) where D between " + from_date + " and " + to_date + " group by interval, L order by interval),users_master where user = users_master.B;"
    user_info_2 = "select url_master.A as urls, users_master.A as users, count(N) as rq, sum(G) as bw from (select N,L,G,D from main where F NOT IN (503,500,407,403) and  D between " + from_date + " and " + to_date + "),users_master,url_master where L=users_master.B and N=url_master.B group by N,L order by bw DESC;"
    user_info_3 = "select url_master.A as urls, users_master.A as users, count(N) as rq, sum(G) as bw from (select N,L,G,D from main where F=403 and D between " + from_date + " and " + to_date + "),users_master,url_master where L=users_master.B and N=url_master.B group by N,L order by bw DESC;"    
    peak_based_bandwidth = "select strftime('%d',datetime(D, 'unixepoch', 'localtime')) as interval, count(D) as rq, sum(G) as bw from (select D,G from main) where D between " + from_date + " and " + to_date + " group by interval order by bw DESC;"									
    peak_based_requests = "select strftime('%d',datetime(D, 'unixepoch', 'localtime')) as interval, count(D) as rq, sum(G) as bw from (select D,G from main) where D between " + from_date + " and " + to_date + " group by interval order by rq DESC;"				
    domain_grouped = "select interval, req_domains_master.A as domains, rq, bw from (select strftime('%H',datetime(D,'unixepoch','localtime')) as interval, Y as domain, count(Y) as rq,sum(G) as bw from (select D,Y,G from main order by D) where D between " + from_date + " and " + to_date + " group by interval,Y order by interval),req_domains_master where domain = req_domains_master.B;"
    hour_analysis = "select interval, users_master.A as users, rq, bw from (select strftime('%H',datetime(D, 'unixepoch', 'localtime')) as interval, L as user, count(L) as rq,sum(G) as bw from (select D,L,G from main order by D) where D between " + from_date + " and " + to_date + " group by interval, L order by interval),users_master where user = users_master.B;"
    denied = "select url_master.A as urls, rq, bw from(select N, count(N) as rq, sum(G) as bw from main where F=403 and D between " + from_date + " and " + to_date + " group by N),url_master where N=url_master.B order by bw DESC;"
    allowed = "select url_master.A as urls, rq, bw from(select N, count(N) as rq, sum(G) as bw from main where F NOT IN (503,500,407,403) and D between " + from_date + " and " + to_date + " group by N),url_master where N=url_master.B order by bw DESC;"
    
    queries = [[global_1, "global_1"], [global_2, "global_2"], [global_3, "global_3"], [userlist, "userlist"], [sitelist, "sitelist"], [user_details_1, "user_details_1"], [user_details_2, "user_details_2"], [user_details_3, "user_details_3"], [user_info_1, "user_info_1"], [user_info_2, "user_info_2"], [user_info_3, "user_info_3"], [peak_based_bandwidth, "peak_based_bandwidth"], [peak_based_requests, "peak_based_requests"], [domain_grouped, "domain_grouped"], [hour_analysis, "hour_analysis"], [denied, "denied"], [allowed, "allowed"]]

    start = time.time()
    threads = []
    for url in queries:
        t = GetQueryThread(url)
        threads.append(t)
        t.start()
    for t in threads:
        t.join()
    logger.info('Complted process in : %s', (time.time()-start))
 
if __name__ == '__main__':        
    date_arguments()    
    logger.info('Generating report from %s [%s]', db_resource, getSize(db_resource))
    container_json["date"] = {'from': int(from_date), 'to': int(to_date)}
    
    check_directory(report_path)
    get_responses()
    
    #computed array which need to be adjusted to JSON
    json_container = json.dumps(data_container)
    load_json = json.loads(json_container)

    global_1_str = config.get('structure_map', 'global_1').split(",")    
    user_details_1_str = config.get('structure_map', 'user_details_1').split(",")

    logger.info('json tag:[global] generating')
    #global tag
    init_container = {}
    for s in global_1_str:            
        init_container[s] = load_json["global_1"][0][s]

    init_container["bls"] = load_json["global_2"][0]["bls"]
    init_container["brs"] = load_json["global_3"][0]["brs"]
    
    container_json["global"] = init_container
    logger.info('json tag:[global] completed')

    logger.info('json tag:[users] generating')    
    init_container = {}
    cnt = 1
    for s in load_json["userlist"]:
        user_identity = "u" + str(cnt)
        init_container[user_identity] = s["users"]  
        user_map[s["users"]] = user_identity
        cnt += 1
        
    container_json["users"] = init_container    
    write_file(report_path + "/users.json", json.dumps(init_container))
    logger.info('json tag:[users] completed')

    logger.info('json tag:[sites] generating')
    init_container = {}
    cnt = 1
    for s in load_json["sitelist"]:
        site_identity = "s" + str(cnt)
        init_container[site_identity] = s["urls"]      
        site_map[s["urls"]] = site_identity
        cnt += 1

    write_file(report_path + "/urls.json", json.dumps(init_container))
    logger.info('json tag:[sites] completed')

    logger.info('json tag:[stat] generating')
    init_container = {}
    for s in load_json["user_details_1"]:        
        sub_init_container = {}
        user_key = user_map[s["users"]]
        for s1 in user_details_1_str:
            sub_init_container[s1] = s[s1]
    
        init_container[user_key] = sub_init_container
        
    container_json["stats"] = init_container 

    for s in load_json["user_details_2"]:            
        user_key = user_map[s["users"]]
        denied = s["dn"] 
        container_json['stats'][user_key]["dn"] = denied    
        container_json['stats'][user_key]["denied"] = []
#	container_json['stats'][user_key]["hour"] = []

    for s in load_json["user_details_3"]:        
        user_key = user_map[s["users"]]
        allow = s["br"]                
        container_json['stats'][user_key]["br"] = allow
        container_json['stats'][user_key]["allowed"] = []
        container_json['stats'][user_key]["hour"] = []

    for s in load_json["user_info_2"]:                
        user_key = user_map[s["users"]]            
        container_json['stats'][user_key]["allowed"].append({'url':site_map[s["urls"]], 'bw': s["bw"], 'rq': s["rq"]})
   
    for s in load_json["user_info_3"]:                
        user_key = user_map[s["users"]]            
        container_json['stats'][user_key]["denied"].append({'url':site_map[s["urls"]], 'bw': s["bw"], 'rq': s["rq"]})

    init_container = []   
    for s in load_json["denied"]:     
        init_container.append({'st': site_map[s["urls"]], 'rq': s["rq"], 'bw': s["bw"]})

    container_json['denied'] = init_container

    init_container = []   
    for s in load_json["allowed"]:                      
        init_container.append({'st': site_map[s["urls"]], 'rq': s["rq"], 'bw': s["bw"]})

    container_json['allowed'] = init_container
    del init_container

    for s in load_json["hour_analysis"]:         
        user_key = user_map[s["users"]]                
        if (user_key):
			try:
			   container_json['stats'][user_key]["hour"]
			except KeyError:
			   container_json['stats'][user_key]["hour"] = []

			logger.info('json user info hour analysis:[user_key]')
			logger.info([user_key])
			container_json['stats'][user_key]["hour"].append({'int':s["interval"], 'bw': s["bw"], 'rq': s["rq"]})            
        
    logger.info('json tag:[stat] completed')

    json_content = json.dumps(container_json)
    write_file(report_path + "/report.json", json_content)
